/*
 * $Id: mrt.c,v 1.20 1998/07/20 01:21:02 labovit Exp $
 */

#include <string.h>
#include <time.h>
#include <signal.h>

#include <mrt.h>
#include <select.h>
#include <trace.h>
#include <timer.h>

mrt_t *MRT;

void	mrt_process_signal ();
int	show_threads (uii_connection_t *uii);
int	show_version (uii_connection_t *uii);
void    show_sockets (uii_connection_t *uii);
void    terminal_monitor (uii_connection_t *uii);
void    no_terminal_monitor (uii_connection_t *uii);


/* init_mrt
 * initialize MRT bookeeping structures. *ALL* MRT programs must call this 
 * routine.
 */
int init_mrt (trace_t *tr) {
   mrt_thread_t *mrt_thread;

   assert (MRT == NULL);
#ifdef HAVE_THR_SETCONCURRENCY
   /* NOTE: Solaris Thread, not Pthread */
   thr_setconcurrency (40);
#endif /* HAVE_THR_SETCONCURRENCY */

   signal (SIGPIPE, mrt_process_signal);
   signal (SIGHUP, mrt_process_signal);
   signal (SIGINT, mrt_process_signal);
   
   MRT = New (mrt_t);
   MRT->start_time = time (NULL);
   MRT->ll_threads = LL_Create (0);
   MRT->ll_trace = LL_Create (0);
   MRT->ll_signal_call_fn = LL_Create (0);
   MRT->ll_gateways = LL_Create (LL_DestroyFunction, destroy_gateway, 0);
   MRT->trace = tr;
   MRT->config_file_name = NULL;

   pthread_mutex_init (&MRT->mutex_lock, NULL);
   init_schedules (tr);

   mrt_thread = New (mrt_thread_t);
   memcpy (mrt_thread->name, "MAIN thread", 11);
   mrt_thread->thread = pthread_self ();
   LL_Add (MRT->ll_threads, mrt_thread);

   MRT->threadn = 10; /* only used on non-thread capable machines */

   init_uii (tr);
   init_timer (tr);
   init_select (tr);
#ifdef notdef
   /* some commands don't need information on own interfaces */
   init_interfaces (tr);
#endif

   /* add some user interface hooks */
   uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show threads", 
		     (void *) show_threads,
		     "Show status of threads");
   uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show version", 
		     (void *) show_version,
		     "Show version and uptime");
   uii_add_command2 (UII_NORMAL, COMMAND_NODISPLAY, "show sockets", 
		     (void *) show_sockets,
		     "Show sockets");

   uii_add_command2 (UII_NORMAL, COMMAND_NODISPLAY, "terminal monitor", 
		     (void *) terminal_monitor,
		     "Set logging information to UII");
   uii_add_command2 (UII_NORMAL, COMMAND_NODISPLAY, "no terminal monitor", 
		     (void *) no_terminal_monitor,
		     "Turn off logging information to UII");

   uii_add_command2 (UII_NORMAL, COMMAND_NODISPLAY, "help", 
		     (void *) uii_show_help,
		     "Description of the interactive help system");
  return (1);
}


void
mrt_exit (status)
    int status;
{
    trace (INFO, MRT->trace, "exit (%d)\n", status);
    exit (status);
}
  

/* mrt_thread_create
 * a wrapper around pthread create, but also adds
 * thread to bookkeeping structures
 * Note: not locking down structure because (for moment) only main
 *       thread calls this routine. Maybe this will change.
 */
#ifdef notdef /* old code */
mrt_thread_t *mrt_thread_create (char *name, schedule_t *schedule,
				 thread_fn_t call_fn, void *arg) 
{
#ifdef HAVE_LIBPTHREAD
  static pthread_attr_t  tattr = { NULL };
  int rc, ret;
#endif /* HAVE_LIBPTHREAD */
  mrt_thread_t *mrt_thread;
  pthread_t thread = 0;

  mrt_thread = New (mrt_thread_t);
  strncpy (mrt_thread->name, name, sizeof (mrt_thread->name) - 1);
  mrt_thread->name [sizeof (mrt_thread->name) - 1] = '\0';

#ifdef HAVE_LIBPTHREAD
#ifdef SOLARIS26
  /* we know this is solaris dependent way -- masaki */
  if (tattr.__pthread_attrp != NULL)
    rc = pthread_attr_destroy (&tattr);
#else
  if (tattr.pthread_attrp != NULL)
    rc = pthread_attr_destroy (&tattr);
#endif /* SOLARIS26 */
  /*pthread_attr_destroy (&tattr);  */
  ret = pthread_attr_init(&tattr);
  ret = pthread_attr_setscope(&tattr, PTHREAD_SCOPE_SYSTEM);
  ret = pthread_attr_setdetachstate (&tattr, PTHREAD_CREATE_DETACHED);
  if (pthread_create (&thread, &tattr, (thread_fn_t) call_fn, arg) < 0) {
    trace (ERROR, MRT->trace, "ERROR -- Could not start %s thread: %s\n",
	   name, strerror (errno));
    return (NULL);
  }

#else
  thread = MRT->threadn++;
  set_thread_id (thread);
  if (schedule)
     schedule->self = thread;
  (*call_fn) (arg);
#endif /* HAVE_LIBPTHREAD */

  mrt_thread->thread = thread;
  mrt_thread->schedule = schedule;
  LL_Add (MRT->ll_threads, mrt_thread);
  return (mrt_thread);
}

#else /* new code */

/* I'm chaging a way to handle pthread_attr being passed to pthread_create ().
   man page doesn't say if I can destroy it right after pthread_create (),
   so I'm saving it into mrt_thread_t structure to delete it when exiting. */
/* This code is important so that I keep the older code for later reference */

mrt_thread_t *mrt_thread_create (char *name, schedule_t *schedule,
				 thread_fn_t call_fn, void *arg) 
{
  mrt_thread_t *mrt_thread;
  pthread_t thread;
  pthread_attr_t tattr;

  /* Locking is required to avoid a confusion 
     in case a created thread executes mrt_thread_exit()
     before adding a mrt_thread structure. */
  pthread_mutex_lock (&MRT->mutex_lock);

#ifdef HAVE_LIBPTHREAD
  pthread_attr_init (&tattr);
  /* as long as I remember, the following attributes are required
     to reuse a thread id. */
  pthread_attr_setscope (&tattr, PTHREAD_SCOPE_SYSTEM);
  pthread_attr_setdetachstate (&tattr, PTHREAD_CREATE_DETACHED);

  if (pthread_create (&thread, &tattr, (thread_fn_t) call_fn, arg) < 0) {
    trace (ERROR, MRT->trace, "ERROR -- Could not start %s thread: %s\n",
	   name, strerror (errno));
    pthread_mutex_unlock (&MRT->mutex_lock);
    pthread_attr_destroy (&tattr);
    return (NULL);
  }

#else
  thread = MRT->threadn++;
  /* a call should be delayed at the end */
#endif /* HAVE_LIBPTHREAD */

  mrt_thread = New (mrt_thread_t);
  /* this expects New () clears the memory */
  if (name)
    strncpy (mrt_thread->name, name, sizeof (mrt_thread->name) - 1);
  mrt_thread->thread = thread;
  mrt_thread->schedule = schedule;
  mrt_thread->attr = tattr;
  LL_Add (MRT->ll_threads, mrt_thread);
  pthread_mutex_unlock (&MRT->mutex_lock);

#ifndef HAVE_LIBPTHREAD
  if (schedule)
     schedule->self = thread;
  set_thread_id (thread);
  (*call_fn) (arg);
#endif /* HAVE_LIBPTHREAD */
  return (mrt_thread);
}

#endif

/*
 * mrt_thread_exit
 */
void mrt_thread_exit () {
  mrt_thread_t *mrt_thread;
  pthread_t thread = pthread_self ();

  assert (thread > 0);
  pthread_mutex_lock (&MRT->mutex_lock);
  LL_Iterate (MRT->ll_threads, mrt_thread) {
     if (mrt_thread->thread == thread) {
        LL_Remove (MRT->ll_threads, mrt_thread);
#ifdef HAVE_LIBPTHREAD
  	pthread_attr_destroy (&mrt_thread->attr);
#endif /* HAVE_LIBPTHREAD */
        Delete (mrt_thread);
	break;
     }
  }
  pthread_mutex_unlock (&MRT->mutex_lock);

#ifdef HAVE_LIBPTHREAD
  pthread_exit (NULL);
#endif /* HAVE_LIBPTHREAD */
}


/* show_version
 * Display version number. Usually called by UII
 */
int show_version (uii_connection_t *uii) {
  
  uii_send_data (uii, "%s\r\n", MRT_VERSION);
  uii_send_data (uii, "  Compiled  on %s\r\n", __DATE__);
  uii_send_data (uii, "  UP for %2.2f hours \r\n", 
		 ((float) (time (NULL) - MRT->start_time)) / (60.0*60.0));
  return (1);
}

/* show_threads
 * Display thread and schedule info. Usually called by UII
 */
int show_threads (uii_connection_t *uii) {
  mrt_thread_t *mrt_thread;

#ifndef HAVE_LIBPTHREAD  
  uii_send_data (uii, "*** NO THREAD SUPPORT ***\r\n");
  uii_send_data (uii, "(Running in degraded mode)\r\n");
#endif /* HAVE_LIBPTHREAD */
  uii_send_data (uii, "Unix Process ID is %d\r\n", getpid ());

  uii_send_data (uii, "[id]\t%-25s\t%-15s%s\t%-15s\r\n", "Name", 
		 "Status", "Queue", "Blocked");

  pthread_mutex_lock (&MRT->mutex_lock);
  LL_Iterate (MRT->ll_threads, mrt_thread) {
    uii_send_data (uii, "[%-2d]\t%-25s\t%-15s%d\t%d\r\n",
		   mrt_thread->thread, mrt_thread->name, "Running", 
		   schedule_count (mrt_thread->schedule), 0);
    
  }
  pthread_mutex_unlock (&MRT->mutex_lock);
  return (1);
}


/* mrt_process_signal
 * invoke registered handler routines, and exit on sigint
 */
void mrt_process_signal (int sig) {

  trace (TR_WARN, MRT->trace, "MRT signal (%d) received\n", sig);
  if (sig == SIGINT) mrt_exit (2);
  /* Should we exit in all cases? */
  signal (sig, mrt_process_signal);
}


void
init_mrt_thread_signals ()
{
    sigset_t set;

    sigemptyset (&set);
    sigaddset (&set, SIGALRM);
    sigaddset (&set, SIGHUP);
    sigaddset (&set, SIGINT);
    sigaddset (&set, SIGPIPE);
    pthread_sigmask (SIG_BLOCK, &set, NULL);
}


#ifndef HAVE_LIBPTHREAD
/* use real assert */
#undef assert
#include </usr/include/assert.h>
int
pthread_mutex_init (pthread_mutex_t *mp, const pthread_mutexattr_t *attr)
{
    *mp = 0;
    return (0);
}

int
pthread_mutex_lock (pthread_mutex_t *mp)
{
    assert (*mp == 0);
    *mp = 1;
    return (0);
}

int
pthread_mutex_trylock (pthread_mutex_t *mp)
{
    if (*mp == 0) {
	*mp = 1;
	return (0);
    }
    return (1);
}

int
pthread_mutex_unlock (pthread_mutex_t *mp)
{
    assert (*mp == 1);
    *mp = 0;
    return (0);
}

int
pthread_mutex_destroy (pthread_mutex_t *mp)
{
    return (0);
}

#endif /* HAVE_LIBPTHREAD */


void mrt_main_loop () {
#ifndef HAVE_LIBPTHREAD
  while (1) {
    while (process_all_schedules ());       /* process all first */
    mrt_select ();
    mrt_alarm ();
  }
#else
  /* loop */
  /* XXX we need to implement a mechanism to wait mrt threds */
  while (1) {
    /* main thread catches alarm interrupts by sigwait() */
    mrt_alarm ();
  }
#endif /* HAVE_LIBPTHREAD */
}
